class Associable(object):
    def __init__(self, associated_with = None):
        self.__associated_with = associated_with

    def associated_with(self):
        return self.__associated_with

    def is_associated(self):
        return self.associated_with() != None

    def associate(self, associate_with):
        if not self.is_associated():
            self.__associated_with = associate_with



from .object import Object
from .associable import Associable
from .bus import Bus
class Pin(Object, Associable):
    def __init__(self, name, associate_with = None):
        Object.__init__(self, name)
        Associable.__init__(self, associate_with)

    def is_part_of_bus(self):
        return isinstance(self.associated_with(), Bus)

    def bus(self):
        if self.is_part_of_bus():
            return self.associated_with()
        else:
            return self

    def verilog_port_list(self):
        if self.is_part_of_bus():
            return self.associated_with().verilog_port_list(self)
        else:
            return self.get_object_name()

    def verilog_declare(self):
        if isinstance(self.associated_with(), Bus):
            return self.associated_with().verilog_declare(self)
        else:
            return self.verilog_type() + ' ' + self.get_object_name() + ';'

    def verilog_type(self):
        return 'inout'

    def __lt__(self, other):
        if self.is_part_of_bus():
            sw = self.bus().width()
        else:
            sw = 0
            
        if other.is_part_of_bus():
            ow = other.bus().width()
        else:
            ow = 0

        if (sw!=ow):
            return sw>ow

        return self.get_object_name() < other.get_object_name()



import re
from .object import Object
from .pin import Pin
from .net import Net
from .bus import Bus
from .param import Param
from copy import copy, deepcopy

class Component(Object):
    __global_components = {}

    def all_components():
        return Component.__global_components

    def get_component_by_name(comp_name):
        return Component.__global_components[comp_name]

    def __init__(self, name):
        super(Component, self).__init__(name)

        if name in Component.__global_components:
            raise Exception('component already defined')
        else:
            Component.__global_components[name] = self

        self.__pins = {} #noytzach - performance
        self.__nets = {} #noytzach - performance
        self.__subcomponents = {}
        self.__net_connectivity = {}
        self.__pin_connectivity = {}
        self.__virtual_pins = {}
        self.__params = []
        self.__pins_ordered = []
        self.add_pin_adds_net = True
        self.__dont_uniq = False
        self.__dont_write_verilog = False
        self.__is_physical = False
        self.__is_sequential = False

    def get_pins(self):
        return self.__pins.values()

    def get_pins_ordered(self):
        return self.__pins_ordered


        

    def get_pin(self, pin_name):
        return self.__pins[pin_name]

    def pin_names(self):
        return self.__pins.keys() #noytzach - performance
        #return [x.get_object_name() for x in self.get_pins()]

    def add_pin(self, pin):
        if not isinstance(pin, Pin):
            raise Exception('not a pin')

        if pin.get_object_name() in self.pin_names():
            raise Exception('pin with same name already exists')

        pin.associate(self)
        self.__pins[pin.get_object_name()] = pin

        if self.add_pin_adds_net and not pin.is_part_of_bus():
            n = pin.get_object_name()
            self.add_net(Net(n))
            self.connect(n, n)

        self.__pins_ordered.append(pin)

    def add_pinbus(self, pinbus):
        if not isinstance(pinbus, Bus):
            raise Exception('not a bus')
        # temporarily disable add_pin_adds_net feature so that pinbus is connected to netbus (later, and not to nets that are not part of bus)
        temp = self.add_pin_adds_net
        self.add_pin_adds_net = False
        for p in pinbus.all_bits():
            self.add_pin(p)
        self.add_pin_adds_net = temp
        
        pinbus.associate(self)

        if self.add_pin_adds_net:
            n = pinbus.get_name()
            self.add_netbus(Bus(Net, n, pinbus.width()))
            self.connect_bus(n, n)

    def get_params(self, str = 0):
        if str:
            return (x.get_string() for x in self.__params)
        else:
            return (x for x in self.__params)

    def add_param(self, param):
        if not isinstance(param, Param):
            raise Exception('not a param')
        if any(x.get_object_name() == param.get_object_name() for x in self.get_params()):
           raise Exception(param.get_object_name() + ' is allready exist')
        self.__params.append(param)

    def set_param(self, param_name, value):
        filtered_list = list(x for x in self.get_params() if x.get_object_name()==param_name)
        if len(filtered_list) == 0:
            raise Exception('param ' + param_name + ' does not exist')
        filtered_list[0].set_value(value)

    def get_nets(self):
        return self.__nets.values()

    def get_net(self, net_name):
        return self.__nets[net_name]

    def net_names(self):
        return self.__nets.keys() #noytzach - performance
        #return [x.get_object_name() for x in self.get_nets()]

    def add_net(self, net):
        if not isinstance(net, Net):
            raise Exception('not a net')

        if net.get_object_name() in self.net_names():
            raise Exception('net "'+net.get_object_name()+'" already exists in component')

        net.associate(self)
        self.__nets[net.get_object_name()] = net #noytzach - performance

    def add_netbus(self, netbus):
        if not isinstance(netbus, Bus):
            raise Exception('not a bus')

        for n in netbus.all_bits():
            self.add_net(n)

        netbus.associate(self)

    def get_subcomponents(self):
        return (x for x in self.__subcomponents)

    def get_subcomponents_recursive(self, inclusive=False):
        s = set()
        for c in self.get_subcomponents():
            dev = self.__subcomponents[c]
            s.update(dev.get_subcomponents_recursive())
            s.add(dev)

        if inclusive:
            s.add(self)
        return s

    def get_subcomponents_dict(self):
        return self.__subcomponents

    def component_names(self):
        return [x for x in self.get_subcomponents()]

    def add_component(self, component, instance_name):
        self.add_subcomponent(component, instance_name)

    def add_subcomponent(self, component, instance_name):
        if not isinstance(component, Component):
            raise Exception('not a component')

        if instance_name in self.__subcomponents:
            raise Exception('component ['+instance_name+'] already exists')
        else:
            self.__subcomponents[instance_name] = component

    def connect(self, net_str, pin_str):
        if not isinstance(net_str, str):
            raise Exception('net_str should be a string')

        try:
            net = self.get_net(net_str)
        except:  # todo: add type
            raise Exception('cannot find net [' + net_str + '] in component')

        if not isinstance(pin_str, str):
            raise Exception('pin_str should be a string')

        lpin = pin_str.split('.')
        if len(lpin)==1:  # pin of component
            try:
                pin = self.get_pin(pin_str)
            except:  # todo: add type
                raise Exception('cannot find pin [' + pin_str + '] in component')

        elif len(lpin)==2:  # pin of sub-component 
            inst, inst_pin = lpin # lpin[0]=instance name, lpin[1]=pin name
            try:
                sub = self.__subcomponents[inst]
            except:  # todo: add type
                raise Exception('sub-component [' + inst + '] not found')

            try:
                pin = sub.get_pin(inst_pin)
            except:  # todo: add type
                raise Exception('pin [' + inst_pin + '] not found in sub-component [' + inst + ']')

            if (inst, inst_pin) not in self.__virtual_pins:
                virtual_pin = Pin(inst + '.' + inst_pin)
                self.__virtual_pins[(inst, inst_pin)] = virtual_pin
                self.__virtual_pins[virtual_pin] = pin

                pin = virtual_pin
            else:
                raise Exception('[' + inst + '.' + inst_pin + '] already connected')

        else:  # too deep
            raise Exception('too deep [' + '.'.join(pin) + ']')

        if pin in self.__pin_connectivity:
            raise Exception('pin already connected')

        if net not in self.__net_connectivity:
            self.__net_connectivity[net] = []

        self.__net_connectivity[net].append(pin)
        self.__pin_connectivity[pin] = net





    def disconnect(self, pin_str):
        if not isinstance(pin_str, str):
            raise Exception('pin_str should be a string')

        lpin = pin_str.split('.')
        if len(lpin)==1:  # pin of component
            raise Exception('cannot disconnect primary pin [' + pin_str + '] from net!')

        elif len(lpin)==2:  # pin of sub-component 
            inst, inst_pin = lpin # lpin[0]=instance name, lpin[1]=pin name
            try:
                sub = self.__subcomponents[inst]
            except:  # todo: add type
                raise Exception('sub-component [' + inst + '] not found')

            try:
                pin = sub.get_pin(inst_pin)
            except:  # todo: add type
                raise Exception('pin [' + inst_pin + '] not found in sub-component [' + inst + ']')

            if (inst, inst_pin) in self.__virtual_pins:
                virtual_pin = self.__virtual_pins[(inst, inst_pin)]
                del self.__virtual_pins[(inst, inst_pin)]
                del self.__virtual_pins[virtual_pin]

                pin = virtual_pin
            else:
                raise Exception('[' + inst + '.' + inst_pin + '] not found as connected')

        else:  # too deep
            raise Exception('too deep [' + '.'.join(pin) + ']')

        if pin not in self.__pin_connectivity:
            raise Exception('pin not connected')
        net = self.__pin_connectivity[pin]

        self.__net_connectivity[net].remove(pin)
        del self.__pin_connectivity[pin]

    def disconnect_bus(self, pinbus_str):
        if not isinstance(pinbus_str, str):
            raise Exception('pinbus_str should be a string')
        
        lpin = pinbus_str.split('.')
        if len(lpin)==1:  # pin of component
            pass # TODO: need to implement this branch
        elif len(lpin)==2: #pin of sub-component
            inst, inst_pin = lpin # lpin[0]=instance name, pin[1]=pin name
            try:
                sub = self.__subcomponents[inst]
            except:  # todo: add type
                raise Exception('sub-component [' + inst + '] not found')

            try:
                pinbus = next(p.bus() for p in sub.get_pins() if p.is_part_of_bus() and p.bus().get_name()==inst_pin)
            except:  # todo: add type
                raise Exception('cannot find pinbus [' + pinbus_str + '] in component')


        for p in pinbus.all_bits():
            self.disconnect(inst+'.'+p.get_object_name())


    def connect_bus(self, netbus_str, pinbus_str):
        if not isinstance(netbus_str, str):
            raise Exception('netbus_str should be a string')

        if not isinstance(pinbus_str, str):
            raise Exception('pinbus_str should be a string')

        try:
            netbus = next(n.bus() for n in self.get_nets() if n.is_part_of_bus() and n.bus().get_name()==netbus_str)
        except:  # todo: add type
            raise Exception('cannot find netbus [' + netbus_str + '] in component')

        lpin = pinbus_str.split('.')
        if len(lpin)==1:  # pin of component
            pinbus = next(p.bus() for p in self.get_pins() if p.is_part_of_bus() and p.bus().get_name()==lpin[0])
            netbus = next(n.bus() for n in self.get_nets() if n.is_part_of_bus() and n.bus().get_name()==lpin[0])

            if netbus.width() == pinbus.width():
                for n, p in zip(netbus.all_bits(), pinbus.all_bits()):
                    self.connect(n.get_object_name(), p.get_object_name())
            else:
                raise Exception('bus widths does not match')
        elif len(lpin)==2: #pin of sub-component
            inst, inst_pin = lpin # lpin[0]=instance name, pin[1]=pin name
            try:
                sub = self.__subcomponents[inst]
            except:  # todo: add type
                raise Exception('sub-component [' + inst + '] not found')

            try:
                pinbus = next(p.bus() for p in sub.get_pins() if p.is_part_of_bus() and p.bus().get_name()==inst_pin)
            except:  # todo: add type
                raise Exception('cannot find pinbus [' + pinbus_str + '] in component')

            if netbus.width() == pinbus.width():
                for n, p in zip(netbus.all_bits(), pinbus.all_bits()):
                    self.connect(n.get_object_name(), inst + '.' + p.get_object_name())
            else:
                raise Exception('bus widths does not match')

    def count_instances(self):
        #first count the module itself
        count = {self.get_object_name():1}

        #then add the sub-counts from submodules
        for inst in self.get_subcomponents():
            dev = self.__subcomponents[inst]
            sub_count = dev.count_instances()

            for m in sub_count:
                if m in count:
                    count[m] += sub_count[m]
                else:
                    count[m] = sub_count[m]
                    
        return count

    def copy(self, copy_name):
        if copy_name in Component.__global_components:
            raise Exception('component ['+copy_name+'] already exists')

        new_dev = copy(self)
        new_dev.set_object_name(copy_name)
        Component.__global_components[copy_name] = new_dev

        return new_dev

    def deepcopy(self, copy_name):
        if copy_name in Component.__global_components:
            raise Exception('component ['+copy_name+'] already exists')

        new_dev = deepcopy(self)
        new_dev.set_object_name(copy_name)
        Component.__global_components[copy_name] = new_dev

        return new_dev
        

    def set_dont_uniq(self, val):
        self.__dont_uniq = val

    def get_dont_uniq(self):
        return self.__dont_uniq

    def set_dont_write_verilog(self, val):
        self.__dont_write_verilog = val

    def get_dont_write_verilog(self):
        return self.__dont_write_verilog

    def set_is_physical(self, val):
        self.__is_physical = val

    def get_is_physical(self):
        return self.__is_physical

    def set_is_sequential(self, val):
        self.__is_sequential = val

    def get_is_sequential(self):
        return self.__is_sequential

    def uniq(self, count = None, numbering = None):
        #topmost component counts, all others should get count from above
        if not count:
            count = self.count_instances()

        #same as count, the one that counts should setup the numbering
        if not numbering:
            numbering = {}
            for k in count:
                numbering[k] = 0
                #make sure k_# not already used
                while k+"_"+str(numbering[k]) in Component.__global_components:
                    numbering[k] +=1
              
        my_fun = lambda k,v,q=0, p=0: [k, int(v), q, int(p)]  
        for inst in sorted(self.get_subcomponents(), key=lambda t: my_fun(*re.match(r'([^\d]+)(\d+)([^\d]+)(\d+)',t+'0_0').groups())):
            dev = self.__subcomponents[inst]
            devname = dev.get_object_name()
                        
            if count[devname]>1 and not dev.get_dont_uniq():
                new_devname = devname+"_"+str(numbering[devname])
                new_dev = dev.deepcopy(new_devname)
                self.__subcomponents.update({inst: new_dev})

                #the new devname is missing from count dictionary
                count[new_devname] = 1

                #increment numbering to next available number
                while True:
                    numbering[devname] += 1
                    next_devname_uniq = devname+"_"+str(numbering[devname])
                    if next_devname_uniq not in Component.__global_components:
                        break
                    
                dev = new_dev

            dev.uniq(count, numbering)
            
        
            
        

    def verilog_port_list(self):
        return [x for x in [pl.verilog_port_list() for pl in self.get_pins()] if x is not None]


    def write_verilog(self):
        v = []
        mod_def = '\nmodule ' + self.get_object_name() + ' ('
        ports = ', '.join(sorted(self.verilog_port_list()))
        mod_def += ports + ');'
        for ll in Component.line_wrap(mod_def):
            v.append('\n' + ll)

        if self.get_pins():
            v.append('')
            v.append('\t//ports')

        for pin in sorted(self.get_pins()):
            pin_dec = pin.verilog_declare()
            if (pin_dec):
                v.append('\t' + pin_dec)
        

        if self.get_nets():
            v.append('')
            v.append('\t//wires')
        
        for net in sorted(self.get_nets()):
            net_dec = net.verilog_declare()
            if (net_dec):
                v.append('\t' + net_dec)


        if self.get_subcomponents():
            v.append('')
            v.append('\t//instances')


		
        my_fun = lambda k,v,q=0, p=0: [k, int(v), q, int(p)]
        for inst in sorted(self.get_subcomponents(), key=lambda t: my_fun(*re.match(r'([^\d]+)(\d+)([^\d]+)(\d+)',t+'0_0').groups())):
            dev = self.__subcomponents[inst]
            l = '\t' + dev.get_object_name() + ' ' + inst + '('
            ports = []
            for inst_pin in [p for p in dev.get_pins() if not p.is_part_of_bus()]:
                inst_pin_name = inst_pin.get_object_name()
                if (inst, inst_pin_name) in self.__virtual_pins:
                    vpin = self.__virtual_pins[(inst, inst_pin_name)]
                    net = self.__pin_connectivity[vpin]
                    net_name = net.get_object_name()
                else:
                    net_name = ''
                ports.append('.' + inst_pin_name + '(' + net_name + ')')

            for inst_bus in set([p.bus() for p in dev.get_pins() if p.is_part_of_bus()]):
                inst_bus_name = inst_bus.get_name()
                concat = []
                for inst_pin in inst_bus.all_bits():
                    inst_pin_name = inst_pin.get_object_name()
                    if (inst, inst_pin_name) in self.__virtual_pins:
                        vpin = self.__virtual_pins[(inst, inst_pin_name)]
                        net = self.__pin_connectivity[vpin]
                        concat.append(net.get_object_name())
                    else:
                        concat.append("1'bx")
                concat = Component.minimize_concat(concat)
                ports.append('.' + inst_bus_name + '({' + ', '.join(concat) + '})')


            l += (', '.join(sorted(ports)))
            l += ');'
            for ll in Component.line_wrap(l):
                v.append(ll)


        v.append('')
        v.append('endmodule')
        return v

    def pin_order(self, x):
        return x

    def write_netlist(self):
        nl = []

        #subcircuits
        for sub in self.get_subcomponents_recursive():
            nl.append('\n// Cell name: ' + sub.get_object_name() + '\n// View name: schematic')

            subckt_def = 'subckt ' + sub.get_object_name() + ' '
            pins = [p.get_object_name() for p in sub.get_pins_ordered()]
            ports = ' '.join(pins)
            subckt_def += ports 
            nl.append(subckt_def)
    
            #instances
            for inst in sorted(sub.get_subcomponents()):
                dev = sub.__subcomponents[inst]
                l = '\t' + inst + ' ('
                ports = []
                for inst_pin in [p for p in dev.get_pins() if not p.is_part_of_bus()]:
                    inst_pin_name = inst_pin.get_object_name()
                    if (inst, inst_pin_name) in sub.__virtual_pins:
                        vpin = sub.__virtual_pins[(inst, inst_pin_name)]
                        net = sub.__pin_connectivity[vpin]
                        net_name = net.get_object_name()
                    else:
                        net_name = ''
                    ports.append(net_name)
    
    
                l += (' '.join(ports))
                l += ') ' + dev.get_object_name()
                param = dev.get_params(1)
                l += ' ' + (' '.join(param))
                for ll in Component.line_wrap(l, endl=' \\'):
                    nl.append(ll)
           
            nl.append('ends ' + sub.get_object_name())
            nl.append('// End of subcircuit definition')
        
	#main
        nl.append('\n// Cell name: ' + self.get_object_name() + '\n// View name: schematic')
    
        my_fun = lambda k,v: [k, int(v)]
        for inst in sorted(self.get_subcomponents(), key=lambda t: my_fun(*re.match(r'([a-zA-Z_]+)(\d+)',t).groups())):
            dev = self.__subcomponents[inst]
            l = inst + ' ('
            ports = []
            for inst_pin in [p for p in dev.get_pins_ordered() if not p.is_part_of_bus()]:
                inst_pin_name = inst_pin.get_object_name()
                if (inst, inst_pin_name) in self.__virtual_pins:
                    vpin = self.__virtual_pins[(inst, inst_pin_name)]
                    net = self.__pin_connectivity[vpin]
                    net_name = net.get_object_name()
                else:
                    net_name = ''
                ports.append(net_name)


            l += (' '.join(ports))
            l += ') ' + dev.get_object_name()
            param = dev.get_params(1)
            l += ' ' + (' '.join(param))
            for ll in Component.line_wrap(l, endl=' \\'):
                nl.append(ll)	

        return nl

    def line_wrap(line, line_width = 150, endl=''):
        """splits a long string to multiple strings (lines), returns a list of strings
        line_width is the MINIMUM number of charechters for line to be split,
        i.e. the line break is added at the first blank charachter AFTER line_width charachters
        endl is added at the line-break - needed for tools that require some escape sequence for multi line command (e.g. \ in tcl)
        """
        lines = []
        i = 0
        curr = ''
        for c in line:
            i = i + 1
            if i>line_width and c.isspace():
                lines.append(curr + endl)
                curr ='\t'
                i = 4
            else:
                curr += c

        if i>0:
            lines.append(curr)

        return lines

    @staticmethod
    def minimize_concat(concat):
        minimized = True
        bit_re = re.compile(r"([0-9]*)'b([01xz]*),([0-9]*)'b([01xz]*)")
        net_re = re.compile(r"(.*)\[([0-9]*:)?([0-9]*)\]\1\[([0-9]*)\]")
        while minimized:
            minimized = False
            for i in range(len(concat)-1):
                m = re.match(bit_re, concat[i] + ',' + concat[i + 1])
                if m:
                    concat[i] = str(int(m.group(1)) + int(m.group(3))) + "'b" + m.group(2) + m.group(4)
                    del concat[i + 1]
                    minimized = True
                    break
                
                m = re.match(net_re, concat[i] + concat[i + 1])
                if m and (int(m.group(3))==(int(m.group(4))+1)):
                    if (m.group(2)):
                        concat[i] = m.group(1) + '[' + m.group(2) + m.group(4) + ']'
                    else:
                        concat[i] = m.group(1) + '[' + m.group(3) + ':' + m.group(4) + ']'
                    del concat[i + 1]
                    minimized = True
                    break


        return concat

    def all_connected(self, net_str):
        connected = []
        if not isinstance(net_str, str):
            raise Exception('net_str should be a string')
        lnet = net_str.replace('.','/').split('/',1)
        if len(lnet)==1:  # net of component
            try:
                net = self.get_net(net_str)
            except:  # todo: add type
                raise Exception('cannot find net [' + net_str + '] in component [' + self.get_object_name() + ']')
            pin_list = self.__net_connectivity[net]
            i = 0
            while i < len(pin_list):
                pin_str = pin_list[i].get_object_name()
                i+=1
                lpin = pin_str.split('.')
                if len(lpin)==1:  # pin of component
                    connected.append(pin_str)
                elif len(lpin)==2:  # pin of sub-component 
                    inst, inst_pin = lpin # lpin[0]=instance name, lpin[1]=pin name
                    sub = self.__subcomponents[inst]
                    if sub.__is_physical:
                        connected.append(pin_str)
                    else:
                        newconnections = sub.all_connected(sub.__pin_connectivity[sub.get_pin(inst_pin)].get_object_name())
                        for p in newconnections:
                            if len(p.split('.'))==1:
                                pin_list.extend([x for x in self.__net_connectivity[self.__pin_connectivity[self.__virtual_pins[(inst,p)]]] if x not in pin_list])
                            else:
                                connected.append(inst + '/' + p)
        else:  # net of a nested subcomponent
            inst, inst_net = lnet # lnet[0]=instance name, lnet[1]=net name
            try:
                sub = self.__subcomponents[inst]
            except:  # todo: add type
                raise Exception('cannot find subcomponent [' + inst + '] in component [' + self.get_object_name() + ']')
            for pin_str in sub.all_connected(inst_net):
                newconnections = [inst+'/'+pin_str]
                if len(pin_str.split('.'))==1:
                    newconnections = self.all_connected(self.__pin_connectivity[self.__virtual_pins[(inst,pin_str)]].get_object_name())
                connected.extend(x for x in newconnections if x not in connected)
        return connected

    '''
0 - full path +dupes
1 - full path 1 per endpoint
2 - internal pins +dupes
3 - internal pins -dupes
4 - endpoints +dupes
5 - endpoints -dupes
    '''

    def all_fan_in(self, net_str, output_type=0):
        fan_in = []
        drive_pins = 0
        for pin_str in self.all_connected(net_str):
            lpin = pin_str.split('.')
            if len(lpin)==1:  # pin of component
                if self.get_pin(pin_str).verilog_type() == "input":
                    fan_in.append(pin_str)
                    drive_pins+=1
            elif len(lpin)==2:  # pin of sub-component 
                inst_str, inst_pin = lpin # lpin[0]=instance name, lpin[1]=pin name
                linst = inst_str.split('/')
                sub = self
                for inst in linst:
                    sub = sub.__subcomponents[inst]
                pin = sub.get_pin(inst_pin)
                if pin.verilog_type() == "output":
                    if sub.__is_sequential:
                        fan_in.append(pin_str)
                        drive_pins+=1
                    else:
                        for inp in [p for p in sub.get_pins() if p.verilog_type() == "input"]:
                            fan_in.extend(x+'>'+pin_str for x in self.all_fan_in(inst_str+'.'+sub.__pin_connectivity[inp].get_object_name()))
            if drive_pins > 1:
                print("WARNING: There are "+str(drive_pins)+" pins driving net "+net_str)
        if output_type==0:
            return fan_in
        elif output_type==1:
            temp = {}
            for p in fan_in:
                temp[p.partition('>')[0]] = p
            return list(temp.values())
        elif output_type==2:
            return [pin for path in [p.split('>') for p in fan_in] for pin in path]
        elif output_type==3:
            return list(set(pin for path in [p.split('>') for p in fan_in] for pin in path))
        elif output_type==4:
            return [p.partition('>')[0] for p in fan_in]
        elif output_type==5:
            return list(set(p.partition('>')[0] for p in fan_in))

    def all_fan_out(self, net_str, output_type=0):
        fan_out = []
        for pin_str in self.all_connected(net_str):
            lpin = pin_str.split('.')
            if len(lpin)==1:  # pin of component
                if self.get_pin(pin_str).verilog_type() == "output":
                    fan_out.append(pin_str)
            elif len(lpin)==2:  # pin of sub-component 
                inst_str, inst_pin = lpin # lpin[0]=instance name, lpin[1]=pin name
                linst = inst_str.split('/')
                sub = self
                for inst in linst:
                    sub = sub.__subcomponents[inst]
                pin = sub.get_pin(inst_pin)
                if pin.verilog_type() == "input":
                    if sub.__is_sequential:
                        fan_out.append(pin_str)
                    else:
                        for inp in [p for p in sub.get_pins() if p.verilog_type() == "output"]:
                            fan_out.extend(pin_str+'>'+x for x in self.all_fan_out(inst_str+'.'+sub.__pin_connectivity[inp].get_object_name()))
        if output_type==0:
            return fan_out
        elif output_type==1:
            temp = {}
            for p in fan_out:
                temp[p.rpartition('>')[-1]] = p
            return list(temp.values())
        elif output_type==2:
            return [pin for path in [p.split('>') for p in fan_out] for pin in path]
        elif output_type==3:
            return list(set(pin for path in [p.split('>') for p in fan_out] for pin in path))
        elif output_type==4:
            return [p.rpartition('>')[-1] for p in fan_out]
        elif output_type==5:
            return list(set(p.rpartition('>')[-1] for p in fan_out))



import re
from .object import Object
from .bus import Bus
from .param import Param
from copy import copy, deepcopy

class Group(Object):
    __global_groups = {}


    def get_group_by_name(group_name):
        return Group.__global_groups[group_name]

    def __init__(self, name):
        super(Group, self).__init__(name)

        if name in Group.__global_groups:
            raise Exception('group already defined')
        else:
            Group.__global_groups[name] = self

        self.__subcomponents = {}
        self.__place_type = 'region'    # guide | region | fence
        self.__position = (0.0 0.0)
        self.__width = 0.0
        self.__hight = 0.0


    def set_group_dimensions(self, width, hight):
        self.__width = width
        self.__hight = hight

    def get_group_dimensions(self):
        width = self.__width
        hight = self.__hight
        return (width, hight)

    def get_position(self):
        return self.__positions

    def get_place_type(self):
        return self.__place_type


    def write_tcl_group_definitions_commands(self):
        l = []
        create_group_command = 'createInstGroup'
        add_inst_to_group_command = 'addInstToInstGroup'
        llx = self.get_position()[0]
        lly = self.get_position()[1]
        urx = llx + self.set_group_dimensions()[0]
        ury = lly + self.set_group_dimensions()[1]
        group_box = str(llx)+' '+str(lly)+' 'str(urx)+' '+str(ury)
        l += [create_group_command+' '+self.get_object_name()+' -'+self.get_place_type()+' '+group_box]

        for inst in self.__subcomponents:
            l += [add_inst_to_group_command+' '+self.get_object_name()+' '+inst]

        return l



class Object(object):
    def __init__(self, name):
        self.__name = name

    def get_object_name(self):
        return self.__name

    def set_object_name(self, name):
        self.__name = name

    def __str__(self):
        return self.__name




#import re
# /data/tsmc/65LP/dig_libs/ARM_FEONLY/arm/tsmc/cln65lp/sc12_base_rvt/r0p0/verilog/sc12_cln65lp_base_rvt.v
# /data/tsmc/65LP/dig_libs/ARM_FEONLY/arm/tsmc/cln65lp/sc12_base_rvt/r0p0/lef/sc12_cln65lp_base_rvt.lef

# /data/tsmc/65LP/dig_libs/ARM_FEONLY/arm/tsmc/cln65lp/sc9_base_rvt/r0p0/verilog/sc9_cln65lp_base_rvt.v
# /data/tsmc/65LP/dig_libs/ARM_FEONLY/arm/tsmc/cln65lp/sc9_base_rvt/r0p0/lef/sc9_cln65lp_base_rvt.lef


mode = input('Choose base class: 1) Component\t2)ComponentXY\t')

in_name = input('Enter Verilog (.v) file to be read and processed: ')
input_file = open(in_name,'r')
print ('opened file: [ %s ]\n' % in_name)

dev_name = ''
Lefs = {}

if mode == '2' or mode == 'ComponentXY':
	lef_name = input('Enter name of lef file: ')
	lef_file = open(lef_name, 'r')
	print('Reading lef file: [%s]\n' % lef_name)
	for line in lef_file:
		words = line.split()
		if words and words[0] == 'MACRO':
			dev_name = words[1]
		if words and words[0] == 'SIZE':
			Lefs[dev_name] = [ words[1] , words[3] ]
	lef_file.close()


out_name = input('Enter name of output file: ')
output_file = open(out_name, 'w+')




output_file.write('from salamandra import *\n\n')

temp_line = []
temp_name = ''
for line in input_file:
	words = line.split() # split line
	outputs = []
	inputs = []
	if words:
		if words[0] == 'module':
			temp_name = words[1]
			if len(words) == 2:
				output_file.write('class ' + temp_name[:-1] +'(ComponentXY):\n')
				output_file.write('\t__defaultName = \'' + temp_name[:-1] +'\'\n\n')			
			else:
				output_file.write('class ' + temp_name +'(ComponentXY):\n')
				output_file.write('\t__defaultName = \'' + temp_name +'\'\n\n')
		if words[0] == 'endmodule':
			output_file.write('\n')
		if words[0] == 'output':
			temp_line = words
			if not len(temp_line):
				break
		if words[0] == 'input':
			output_file.write('\tdef __init__(self, name = __defaultName, ')
			for k in range(1,len(words) ):
				inputs.append(words[k][:-1])
			for k in range(0,len(inputs) ):
				output_file.write('in'+ str(k+1) + '=\'' + inputs[k]+'\'' )
				if k != len(inputs):
					output_file.write(', ')
				elif len(temp_line) == 0:
					output_file.write('):\n')
				else: 
					output_file.write(', ')
			for k in range(1, len(temp_line) ):
				outputs.append(temp_line[k][:-1] )
			for k in range(0,len(outputs) ):
				output_file.write('out' + str(k+1) + '=\''+ outputs[k] + '\'')
				if k != len(outputs)-1:
					output_file.write(', ')
				else:
					output_file.write('):\n')
			if mode == '2' or mode == 'ComponentXY':
				output_file.write('\t\tComponentXY.__init__(self,name)\n')
			else:
				output_file.write('\t\tComponent.__init__(self,name)\n')
			for k in range(0,len(inputs) ):
				output_file.write('\t\tself.add_pin(Input(in'+ str(k+1) + ') )\n' )
			
			for k in range(0,len(outputs) ):
				output_file.write('\t\tself.add_pin(Output(out'+ str(k+1) + ') )\n' )
			if mode == '2' or mode == 'ComponentXY':
				output_file.write('\t\tself.set_component_dimensions('+Lefs[temp_name][0]+', '+Lefs[temp_name][1]+')\n' )

			output_file.write('\t\tself.set_dont_uniq(True)\n')
			output_file.write('\t\tself.set_dont_write_verilog(True)\n')
			output_file.write('\t\tself.set_is_physical(True)\n')


input_file.close()
output_file.close()
	



"""
Filename:  sclib.py
Author:    Shai Yemini 
written:   April,2018
"""
#This library contains common pre-built standard-cells.
#It is mainly aimed for net-list creation purposes, and as such the cells are built mostly without the inner logic.
#
#The cells are heavily based on (and default to) tcbn65lp standard-cell library by TSMC
#for cell and pin names and structure.
#####################################################################################################################
##USAGE EXAMPLE: Creating a simple, default AND gate##
#####################################################################################################################
'''
from sclib import *

myAnd = And()
'''
#####################################################################################################################
##THATS IT! you can also customize the values and names if you want (see example files).
#####################################################################################################################
#
#Feel free to add more gates\cells\anything if you like,
#but please try and use the template given below when adding a new cell:
#
#####################################################################################################################
##TEMPLATE FOR GATES:
#####################################################################################################################
'''
#Cell description (Optional: name in library)
class T(Component):
    __defaultName = 'Tx'
    
    def __init__(self, name=__defaultName):
        Component.__init__(self,name)
'''
#####################################################################################################################

from salamandra import *
#####################################################################################################################
##Generic Gates
#####################################################################################################################

#Generic gate with 1 input and 1 output
class G1in1out(Component):
    __defaultName = 'G1IN1OUT'
    
    def __init__(self, name=__defaultName, in1='in1', out='out'):
        Component.__init__(self,name)
        self.add_pin(Output(out))
        self.add_pin(Input(in1))


#Generic gate with 2 inputs and 1 output
class G2in1out(G1in1out):
    __defaultName = 'G2IN1OUT'
    
    def __init__(self, name=__defaultName, in1='in1', in2='in2', out='out'):
        G1in1out.__init__(self,name,in1,out)
        self.add_pin(Input(in2))

#Generic gate with variable number of inputs and 1 output
class Gxin1out(Component):
    __defaultName = 'GXIN1OUT'
    
    def __init__(self, name=__defaultName, inNum=1, out='out', inArr=None, nameFormat='in'):
        Component.__init__(self,name)
        self.add_pin(Output(out))
        self.addInputs(inNum,inArr,nameFormat)
        
    
    def addInputs(self, inNum,inArr=None,nameFormat='in'):
        if (inArr is None):
            inArr = [nameFormat+'1']
            
        arrSize = len(inArr)
        for i in range(1,inNum+1):
            nextIn=''            
            if (i<=arrSize):
                nextIn=inArr[i-1]
            else:
                nextIn=nameFormat+str(i)
            self.add_pin(Input(nextIn))


#Generic gate with variable number of inputs and outputs
class Gxinxout(Gxin1out):
    __defaultName = 'GXINXOUT'
    
    def __init__(self, name=__defaultName, inNum=1, inArr=None, inNameFormat='in', outNum=1, outArr=None, outNameFormat='out'):
        Component.__init__(self,name)
        self.addInputs(inNum,inArr,inNameFormat)
        self.addOutputs(outNum,outArr,outNameFormat)
        
    
    def addOutputs(self,outNum,outArr=None,nameFormat='out'):
        if (outArr is None):
            outArr = [nameFormat+'1']
            
        arrSize = len(outArr)
        for i in range(1,outNum+1):
            nextOut=''            
            if (i<=arrSize):
                nextOut=outArr[i-1]
            else:
                nextOut=nameFormat+str(i)
            self.add_pin(Output(nextOut))
            
#####################################################################################################################
##Transistors (not standard cells) 
#####################################################################################################################
            
#Nmos
class Nmos(Component):
    __defaultName = 'Nmos'
    
    def __init__(self, name=__defaultName, source='source', drain='drain',gateIn='gate',bulk='bulk'):
        Component.__init__(self,name)
        self.add_pin(Inout(source))
        self.add_pin(Inout(drain))
        self.add_pin(Input(gateIn))
        self.add_pin(Input(bulk))
   
#Pmos
class Pmos(Nmos):
    __defaultName = 'Pmos'
    
    def __init__(self, name=__defaultName, source='source', drain='drain',gateIn='gate',bulk='bulk'):
        Nmos.__init__(self, name,source,drain,gateIn,bulk)

#####################################################################################################################
## Misc.
#####################################################################################################################
        
#Inverter (INVDx)
class Inverter(G1in1out):
    __defaultName = 'INVD0'
    
    def __init__(self,name=__defaultName, in1='I', out='ZN'):
        G1in1out.__init__(self,name,in1,out)

#Buffer (BUFFDx)
class Buffer(G1in1out):
    __defaultName = 'BUFFD0'
    
    def __init__(self,name=__defaultName, in1='I', out='Z'):
        G1in1out.__init__(self,name,in1,out)
        
#Delay (DELx)
class Delay(G1in1out):
    __defaultName = 'DEL0'
    
    def __init__(self,name=__defaultName, in1='I', out='Z'):
        G1in1out.__init__(self,name,in1,out)
        
#D-FF (DFDx)
class Dff(G2in1out):
    __defaultName = 'DFD0'
    
    def __init__(self, name=__defaultName, in1='D', in2='CP', out1='Q',out2='QN'):
        G2in1out.__init__(self,name,in1,in2,out1)
        self.add_pin(Output(out2))

#####################################################################################################################
##AND Gates
#####################################################################################################################
        
#AND2 (AN2Dx)
class And(G2in1out):
    __defaultName = 'AN2D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', out='Z'):
        G2in1out.__init__(self,name,in1,in2,out)


#AND3 (AN3Dx)
class And3(And):
    __defaultName = 'AN3D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', out='Z'):
        And.__init__(self,name,in1,in2,out)
        self.add_pin(Input(in3))


#AND4 (AN4Dx)
class And4(And3):
    __defaultName = 'AN4D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', in4='A4', out='Z'):
        And3.__init__(self,name,in1,in2,in3,out)
        self.add_pin(Input(in4))
        

#Flexible AND
class Andx(Gxin1out):
    __defaultName = 'AN2D0'
    
    def __init__(self, name=__defaultName, inNum=2, out='Z', inArr=None, nameFormat='A'):
        Gxin1out.__init__(self,name,inNum,out,inArr,nameFormat)

#####################################################################################################################
##OR Gates
#####################################################################################################################
        
#OR2 (OR2Dx)
class Or(G2in1out):
    __defaultName = 'OR2D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', out='Z'):
        G2in1out.__init__(self,name,in1,in2,out)


#OR3 (OR3Dx)
class Or3(Or):
    __defaultName = 'OR3D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', out='Z'):
        Or.__init__(self,name,in1,in2,out)
        self.add_pin(Input(in3))


#OR4 (OR4Dx)
class Or4(Or3):
    __defaultName = 'OR4D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', in4='A4', out='Z'):
        Or3.__init__(self,name,in1,in2,in3,out)
        self.add_pin(Input(in4))


#Flexible OR
class Orx(Gxin1out):
    __defaultName = 'OR2D0'
    
    def __init__(self, name=__defaultName, inNum=2, out='Z', inArr=None, nameFormat='A'):
        Gxin1out.__init__(self,name,inNum,out,inArr,nameFormat)

#####################################################################################################################
##XOR Gates
#####################################################################################################################
        
#XOR2 (XOR2Dx)
class Xor(G2in1out):
    __defaultName = 'XOR2D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', out='Z'):
        G2in1out.__init__(self,name,in1,in2,out)


#XOR3 (XOR3Dx)
class Xor3(Xor):
    __defaultName = 'XOR3D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', out='Z'):
        Xor.__init__(self,name,in1,in2,out)
        self.add_pin(Input(in3))


#XOR4 (XOR4Dx)
class Xor4(Xor3):
    __defaultName = 'XOR4D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', in4='A4', out='Z'):
        Xor3.__init__(self,name,in1,in2,in3,out)
        self.add_pin(Input(in4))


#Flexible XOR
class Xorx(Gxin1out):
    __defaultName = 'XOR2D0'
    
    def __init__(self, name=__defaultName, inNum=2, out='Z', inArr=None, nameFormat='A'):
        Gxin1out.__init__(self,name,inNum,out,inArr,nameFormat)

#####################################################################################################################
##NAND Gates
#####################################################################################################################

#NAND2 (ND2Dx)
class Nand(And):
    __defaultName = 'ND2D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', out='ZN'):
        And.__init__(self,name,in1,in2,out)
        

#NAND3 (ND3Dx)
class Nand3(Nand):
    __defaultName = 'ND3D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', out='ZN'):
        Nand.__init__(self,name,in1,in2,out)
        self.add_pin(Input(in3))


#NAND4 (ND4Dx)
class Nand4(Nand3):
    __defaultName = 'ND4D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', in4='A4', out='ZN'):
        And3.__init__(self,name,in1,in2,in3,out)
        self.add_pin(Input(in4))
        

#Flexible NAND
class Nandx(Gxin1out):
    __defaultName = 'ND2D0'
    
    def __init__(self, name=__defaultName, inNum=2, out='ZN', inArr=None, nameFormat='A'):
        Gxin1out.__init__(self,name,inNum,out,inArr,nameFormat)
        
#####################################################################################################################
##NOR Gates
#####################################################################################################################

#NOR2 (NR2Dx)
class Nor(Or):
    __defaultName = 'NR2D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', out='ZN'):
        Or.__init__(self,name,in1,in2,out)
        

#NOR3 (NR3Dx)
class Nor3(Nor):
    __defaultName = 'NR3D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', out='ZN'):
        Nor.__init__(self,name,in1,in2,out)
        self.add_pin(Input(in3))


#NOR4 (NR4Dx)
class Nor4(Nor3):
    __defaultName = 'NR4D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='A3', in4='A4', out='ZN'):
        Nor3.__init__(self,name,in1,in2,in3,out)
        self.add_pin(Input(in4))
        

#Flexible NOR
class Norx(Gxin1out):
    __defaultName = 'NR2D0'
    
    def __init__(self, name=__defaultName, inNum=2, out='ZN', inArr=None, nameFormat='A'):
        Gxin1out.__init__(self,name,inNum,out,inArr,nameFormat)

#####################################################################################################################
##AO's
#####################################################################################################################
        
#AO21 (AO21Dx)
class Ao21(And3):
    __defaultName = 'AO21D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='B', out='Z', andG=None, orG=None):
        And3.__init__(self,name,in1,in2,in3,out)
        
        if (andG is None):
            try:
                andGate = Component.get_component('AN2D0')
            except:
                andGate = And()
            self.add_component(andGate,'and')
        else:
            self.add_component(And(andG),'and')

        if (orG is None):
            try:
                orGate = Component.get_component('OR2D0')
            except:
                orGate = Or()
            self.add_component(orGate,'or')
        else:
            self.add_component(Or(orG),'or')
            
        self.connect(in1,'and.A1')
        self.connect(in2,'and.A2')
        self.add_net(Net('andZ'))
        self.connect('andZ','and.Z')
        self.connect('andZ','or.A1')
        self.connect(in3,'or.A2')
        self.connect(out,'or.Z')

#AO211 (AO211Dx)
class Ao211(And4):
    __defaultName = 'AO211D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='B', in4='C', out='Z', andG=None, orG=None):
        And4.__init__(self,name,in1,in2,in3,in4,out)
        
        if (andG is None):
            try:    
                andGate = Component.get_component('AN2D0')
            except:
                andGate = And()
            self.add_component(andGate,'and')
        else:
            self.add_component(And(andG),'and')

        if (orG is None):
            try:    
                orGate = Component.get_component('OR3D0')
            except:
                orGate = Or3()
            self.add_component(orGate,'or3')
        else:
            self.add_component(Or3(orG),'or3')
            
        self.connect(in1,'and.A1')
        self.connect(in2,'and.A2')
        self.add_net(Net('andZ'))
        self.connect('andZ','and.Z')
        self.connect('andZ','or3.A1')
        self.connect(in3,'or3.A2')
        self.connect(in4,'or3.A3')
        self.connect(out,'or3.Z')


#AO221 (AO221Dx)
class Ao221(Andx):
    __defaultName = 'AO221D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='B1', in4='B2', in5='C', out='Z', andG=None, orG=None):
        Andx.__init__(self, name, 5, out, inArr=[in1,in2,in3,in4,in5])
        
        if (andG is None):
            try:    
                andGate = Component.get_component('AN2D0')
            except:
                andGate = And()
        else:
            andGate = And(andG)
            
        self.add_component(andGate,'and1')
        self.add_component(andGate,'and2')

        if (orG is None):
            try:    
                orGate = Component.get_component('OR3D0')
            except:
                orGate = Or3()
            self.add_component(orGate,'or3')
        else:
            self.add_component(Or3(orG),'or3')
            
        self.connect(in1,'and1.A1')
        self.connect(in2,'and1.A2')
        self.connect(in3,'and2.A1')
        self.connect(in4,'and2.A2')
        self.add_net(Net('and1Z'))
        self.add_net(Net('and2Z'))
        self.connect('and1Z','and1.Z')
        self.connect('and2Z','and2.Z')
        self.connect('and1Z','or3.A1')
        self.connect('and2Z','or3.A2')
        self.connect(in5,'or3.A3')
        self.connect(out,'or3.Z')

#####################################################################################################################
##AOI's
#####################################################################################################################
        
#AOI21 (AOI21Dx)
class Aoi21(And3):
    __defaultName = 'AOI21D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='B', out='ZN',aoG=None, andG=None, orG=None, invG=None):
        And3.__init__(self,name,in1,in2,in3,out)
        
        if (aoG is None):
            try:
                ao21 = Component.get_component('AO21D0')
            except:
                ao21 = Ao21(andG=andG, orG=orG)
            self.add_component(ao21,'ao21')
        else:
            self.add_component(Ao21(name=aoG,andG=andG,orG=orG),'ao21')
        
        if (invG is None):
            try:    
                inv = Component.get_component('INVD0')
            except:
                inv = Inverter()
            self.add_component(inv,'inv')
        else:
            self.add_component(Inverter(invG),'inv')
        
        self.connect(in1,'ao21.A1')
        self.connect(in2,'ao21.A2')
        self.connect(in3,'ao21.B')
        self.add_net(Net('aoZ'))
        self.connect('aoZ','ao21.Z')
        self.connect('aoZ','inv.I')
        self.connect(out,'inv.ZN')


#AOI211 (AOI211Dx)
class Aoi211(Andx):
    __defaultName = 'AOI221D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='B', in4='C', out='ZN',aoG=None, andG=None, orG=None, invG=None):
        And4.__init__(self,name,in1,in2,in3,in4,out)
        
        if (aoG is None):
            try:    
                ao211 = Component.get_component('AO211D0')
            except:
                ao211 = Ao211(andG=andG, orG=orG)
            self.add_component(ao211,'ao211')
        else:
            self.add_component(Ao211(name=aoG,andG=andG,orG=orG),'ao211')
        
        if (invG is None):
            try:    
                inv = Component.get_component('INVD0')
            except:
                inv = Inverter()
            self.add_component(inv,'inv')
        else:
            self.add_component(Inverter(invG),'inv')
        
        self.connect(in1,'ao211.A1')
        self.connect(in2,'ao211.A2')
        self.connect(in3,'ao211.B')
        self.connect(in4,'ao211.C')
        self.add_net(Net('aoZ'))
        self.connect('aoZ','ao211.Z')
        self.connect('aoZ','inv.I')
        self.connect(out,'inv.ZN')
        
        
#AOI221 (AOI221Dx)
class Aoi221(Andx):
    __defaultName = 'AOI211D0'
    
    def __init__(self, name=__defaultName, in1='A1', in2='A2', in3='B1', in4='B2', in5='C', out='ZN',aoG=None, andG=None, orG=None, invG=None):
        Andx.__init__(self, name, 5, out, inArr=[in1,in2,in3,in4,in5])
        
        if (aoG is None):
            try:    
                ao221 = Component.get_component('AO221D0')
            except:
                ao221 = Ao211(andG=andG, orG=orG)
            self.add_component(ao221,'ao221')
        else:
            self.add_component(Ao211(name=aoG,andG=andG,orG=orG),'ao221')
        
        if (invG is None):
            try:    
                inv = Component.get_component('INVD0')
            except:
                inv = Inverter()
            self.add_component(inv,'inv')
        else:
            self.add_component(Inverter(invG),'inv')
        
        self.connect(in1,'ao221.A1')
        self.connect(in2,'ao221.A2')
        self.connect(in3,'ao221.B1')
        self.connect(in4,'ao221.B2')
        self.connect(in5,'ao221.C')
        self.add_net(Net('aoZ'))
        self.connect('aoZ','ao221.Z')
        self.connect('aoZ','inv.I')
        self.connect(out,'inv.ZN')



from .associable import Associable
class Bus(Associable):
    def __init__(self, bit_type, name, width, associate_with = None):
        Associable.__init__(self, associate_with)
        self.__name = name
        self.__type = bit_type

        if isinstance(width, tuple):
            self.__msb = max(width)
            self.__lsb = min(width)
        elif isinstance(width, int):
            self.__msb = width-1
            self.__lsb = 0
        else:
            raise Exception('type of width unsupported - should be tuple or int')

        self.associate(associate_with)

        self.__bits = [None] * self.width()

        for i in self.indices():
            self.set_bit(i, self.__type(name + '[' + str(i) + ']', self))

    def get_name(self):
        return self.__name

    def msb(self):
        return self.__msb

    def lsb(self):
        return self.__lsb

    def width(self):
        return (self.msb() - self.lsb() + 1)

    def indices(self):
        return range(self.lsb(), self.msb() + 1)

    def set_bit(self, i, obj):
        if obj.is_associated() and obj.associated_with() != self:
            raise Exception('bit should be associated with bus')
        else:
            obj.associate(self)

        if self.__type != type(obj):
            raise Exception('added bit should be of same type as bus')

        self.__bits[i] = obj

    def verilog_type(self):
        return self.get_bit(self.lsb()).verilog_type()

    def verilog_port_list(self, p):
        if p is self.get_bit(self.lsb()):
            return self.__name
        else:
            return None

    def verilog_declare(self, p):
        if p is self.get_bit(self.lsb()):
            return self.verilog_type() + ' [' + str(self.msb()) + ':' + str(self.lsb()) + '] ' + self.__name + ';'
        else:
            return None
        

    def get_bit(self, i):
        return self.__bits[i]

    def all_bits(self):
        return (self.get_bit(i) for i in sorted(self.indices(), reverse=True))



import re
from .object import Object

class Param(Object):
    def __init__(self, name, default_value = 0, is_editable = True):
        Object.__init__(self, name)
        self.__default_value = default_value
        if not isinstance(is_editable, bool):
            raise Exception('is_editable is boolean')
        self.__is_editable = is_editable
        self.value = self.__default_value

    def is_editable(self):
        return self.__is_editable

    def set_value(self, value):
        if not self.__is_editable:
            raise Exception('[' + self.get_object_name() + '] is not an editable param')
        self.value = value

    def get_value(self):
        return self.value

    def __str__(self):
        return self.get_object_name() + ' = ' + str(self.get_value())

    def get_string(self):
        return self.get_object_name() + '=' + str(self.get_value())

    __repr__ = __str__



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
        name="salamandra",
        version="0.1.0",
        author="Tzachi Noy",
        author_email="tzachi.noy@biu.ac.il",
        description="Framework for describing hierarchical systems",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="http://enicslabs.com/",
        packages=setuptools.find_packages(),
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            ],
        python_requires='>=3.5',
        )



from .component import Component
import re
from enum import Enum

class group_type(Enum):
    guide = 1
    region = 2
    fence = 3

class inst_place_type(Enum):
    fixed = 1
    placed = 2


class ComponentXY(Component):
    def __init__(self, name):
        super(ComponentXY, self).__init__(name)

        self.__positions = {}
        self.__pin_positions = {}
        self.__width = 0.0
        self.__hight = 0.0
        self.__group = None

    def add_component(self, component, instance_name, xcoord=0.0, ycoord=0.0, xmin=None, ymin=None, xmax=None, ymax=None, fixed_OR_placed = 'fixed', group_type = None):
        self.add_subcomponent(component, instance_name, xcoord, ycoord, xmin, ymin, xmax, ymax, fixed_OR_placed, group_type)
		

    def add_subcomponent(self, component, instance_name, xcoord=0.0, ycoord=0.0, xmin=None, ymin=None, xmax=None, ymax=None, fixed_OR_placed = 'fixed', group_type = None):
        Component.add_subcomponent(self, component, instance_name)
        ComponentXY.set_position(self, instance_name, xcoord, ycoord, xmin, ymin, xmax, ymax, fixed_OR_placed, group_type)

    def get_position(self, inst):
        return self.__positions[inst]

    def set_position(self, instance_name, xcoord=0.0, ycoord=0.0, xmin=None, ymin=None, xmax=None, ymax=None, fixed_OR_placed = 'fixed', group_type = None):
        if xmin is None: xmin = xcoord
        if xmax is None: xmax = xcoord
        if ymin is None: ymin = ycoord
        if ymax is None: ymax = ycoord
        self.__positions[instance_name] = (xmin, ymin, xmax, ymax, fixed_OR_placed, group_type)
        #print(str(instance_name)+' -> '+str(self.__positions[instance_name]))  # --> for debug

    def set_component_dimensions(self, width, hight):
        self.__width = width
        self.__hight = hight

    def get_component_dimensions(self):
        width = self.__width
        hight = self.__hight
        return (width, hight)

    def calc_component_dimensions(self):	# calculate component dimensions recursive
        width = 0.0
        height = 0.0

        comp_dict = self.get_subcomponents_dict()
        if len(comp_dict) == 0: #dictionary is empty --> leaf cell
            return (self.get_component_dimensions())
        else: #not leaf --> recurse
            for inst, comp in comp_dict.items():
                if isinstance(comp, ComponentXY):
                    pos = self.get_position(inst)
                    x = pos[0]
                    y = pos[1]
                    dim = comp.calc_component_dimensions()
                    w = dim[0]
                    h = dim[1]

                    new_w = x+w
                    new_h = y+h

                    width = max(width, new_w)
                    height = max(height, new_h)	

                else: 
                    raise Exception(str(inst)+' is not ComponentXY')
            self.set_component_dimensions(width, height)
        return (width, height)

    def get_positions_dict(self):
        return self.__positions

    def get_pin_positions_dict(self):
        return self.__pin_positions

    def get_positions_recursive(self, path="", xmin=0.0, ymin=0.0, xmax=0.0, ymax=0.0, fixed_OR_placed='fixed', group_type=None):
        comp_dict = self.get_subcomponents_dict()
        if len(comp_dict) == 0: #dictionary is empty --> leaf cell
            return [(path, xmin, ymin, xmax, ymax, fixed_OR_placed, group_type)]
        else: #not leaf --> recurse
            if path != "":
                path += "/"
            l = []
            for inst, comp in comp_dict.items():
                inst_position = self.get_position(inst)
                if isinstance(comp, ComponentXY):
                    inst_pos_list = comp.get_positions_recursive(path+inst, xmin + inst_position[0], ymin + inst_position[1], xmax + inst_position[2], ymax + inst_position[3], inst_position[4], inst_position[5]) 
                    l += inst_pos_list
                else: 
                    raise Exception(str(inst)+' is not ComponentXY')
            return l

    def set_pin_position(self, pin_name, xcoord = 0.0, ycoord = 0.0, side = 'LEFT', layer = 2):
        self.__pin_positions[pin_name] = {}
        self.__pin_positions[pin_name]['assign'] = (xcoord, ycoord)
        self.__pin_positions[pin_name]['side'] = side
        self.__pin_positions[pin_name]['layer'] = layer

    def write_tcl_placement_commands(self, tool, put_placed_inst_in_comment=0):
        place_instance_list = []
        create_group_list = []
        add_inst_to_group_list = []
        
        groups_dict = {}
        group_index = 0

        if tool == 'Cadence':
            place_command = 'placeInstance'
            for line in self.get_positions_recursive():
                #print(line)
                if line[0]==line[2] or line[1]==line[3]:    # if xmin == xmax or ymin == ymax so regular placeInstance
                    if (line[5] == 'placed'):
                        place_instance_list += [place_command + ' ' + line[0] + ' [expr ($x_coordinate+' + str(round(line[1], 2)) + ')] [expr ($y_coordinate+' + str(round(line[2], 2)) + ')] -placed']
                    elif (line[5] == 'fixed'):
                        place_instance_list += [place_command + ' ' + line[0] + ' [expr ($x_coordinate+' + str(round(line[1], 2)) + ')] [expr ($y_coordinate+' + str(round(line[2], 2)) + ')] -fixed']                
                    else:
                        place_instance_list += [place_command + ' ' + line[0] + ' [expr ($x_coordinate+' + str(round(line[1], 2)) + ')] [expr ($y_coordinate+' + str(round(line[2], 2)) + ')]']

                    if (put_placed_inst_in_comment == 1 and line[5] == 'placed'):
                        place_instance_list[-1] = '# '+place_instance_list[-1]
                else:
                    group_exist = 0
                    for group in groups_dict:
                        if groups_dict[group]['box'] == (line[1],line[2],line[3],line[4]):
                            groups_dict[group]['inst_list'] += [line[0]]
                            group_exist = 1
                    if group_exist == 0:
                        groups_dict[group_index] = {}
                        groups_dict[group_index]['box'] = (line[1],line[2],line[3],line[4])
                        groups_dict[group_index]['inst_list'] = [line[0]]
                        group_index += 1
            print(groups_dict)
            for group in groups_dict:
                create_group_command = 'createInstGroup'
                box = str(groups_dict[group]['box'][0])+' '+str(groups_dict[group]['box'][1])+' '+str(groups_dict[group]['box'][2])+' '+str(groups_dict[group]['box'][3])
                create_group_list += [create_group_command+' '+str(group)+' -fence {'+box+'}']

                for inst in groups_dict[group]['inst_list']:
                    add_inst_to_group_list += ['addInstToInstGroup'+' '+str(group)+' '+inst]
        
        elif tool == 'Synopsys':
            place_command = 'move_objects'
            for line in self.get_positions_recursive():
                if (line[5] == 1):
                    place_instance_list += [place_command + ' [get_flat_cells ' + line[0] + '] -x [expr ($x_coordinate+' + str(round(line[1], 2)) + ')] -y [expr ($y_coordinate+' + str(round(line[2], 2)) + ')] -placed']
                else:
                    place_instance_list += [place_command + ' [get_flat_cells ' + line[0] + '] -x [expr ($x_coordinate+' + str(round(line[1], 2)) + ')] -y [expr ($y_coordinate+' + str(round(line[2], 2)) + ')]']

        return (place_instance_list, create_group_list, add_inst_to_group_list)
		
    def write_pin_tcl_placement_commands(self):
        l = []
        place_command = 'editPin'
        pin_dict = self.get_pin_positions_dict()
        for pin in sorted(pin_dict):
            l += [place_command + ' -pin "' + str(pin) + '" -layer ' + str(pin_dict[pin]['layer']) + ' -side '+str(pin_dict[pin]['side'])+' -assign '+ str(round(pin_dict[pin]['assign'][0],2))+' '+str(round(pin_dict[pin]['assign'][1],2))+' -snap TRACK -fixOverlap']

        return l

    def write_floorPlan_tcl_commands(self,tech_site):
        self.calc_component_dimensions()
        dimensions = self.get_component_dimensions()
        l = 'floorPlan -site '+str(tech_site)+' -adjustToSite -s '+' [expr ($floorPlan_margin_x+' + str(round(dimensions[0],2))+')] [expr ($floorPlan_margin_y+'+str(round(dimensions[1],2))+')] 0 0 0 0'
        return l

    def write_addStripe_tcl_commands(self, welltap_width = 0.4, min_stripe_spacing = 0.1, min_stripe_width = 0.1):
        l = []
        place_command = 'addStripe'
        # Calculations:
        stripe_width = (welltap_width - min_stripe_spacing)/2
        stripe_width = max(stripe_width,min_stripe_width)
        two_stripe_width = min_stripe_spacing + (2*stripe_width)
        print('\n==========+=============')
        dictt =self.get_subcomponents_dict()
        for k in sorted(dictt):
            #if dictt[k].get_object_name() == 'welltap_stripe':
            if re.match('welltap_stripe', dictt[k].get_object_name()):

                x_pos = round(self.get_positions_dict()[k][0],2)
                stripe_area = str(x_pos) + ' 0.0 '+ str(round(x_pos + 0.4,2)) +' '+ str(round(dictt[k].get_component_dimensions()[1],2))
                print(k+'\t->\tstripe_area: '+str(stripe_area))
                l += [place_command + ' -nets {gnd vdd} -layer M2 -direction vertical -width "'+str(round(stripe_width,2))+' '+str(round(stripe_width,2))+'" -spacing '+str(round(min_stripe_spacing,2))+' -number_of_sets 1 -area "'+ stripe_area +'" -start_offset 0']

        return l
		



from .associable import *
from .bus import *
from .component import *
from .componentXY import *
from .inout import *
from .input import *
from .net import *
from .object import *
from .output import *
from .param import *
from .pin import *
from .sclib import *
